from gdmixworkflow.common.constants import *
from subprocess import Popen,PIPE,STDOUT


def get_param_list(params):
    """ transform params from dict to list.
    """
    if isinstance(params, dict):
        kvList = [(str(k), str(v)) for (k, v) in params.items()]
        paramList = [elem for tupl in kvList for elem in tupl]
        return paramList
    else:
        raise ValueError("job params can only be dict")


def get_tfjob_cmd(params):
    """ get tfjob command for local execution
    """
    cmd = ['python', '-m', 'gdmix.gdmix']
    cmd.extend(get_param_list(params))
    return cmd


def get_sparkjob_cmd(class_name, params, jar='gdmix-data-all_2.11.jar'):
    """ get spark command for local execution
    """
    cmd = ['spark-submit',
           '--class', class_name,
           '--master', 'local[*]',
           '--num-executors','1',
           '--driver-memory', '1G',
           '--executor-memory', '1G',
           '--conf', 'spark.sql.avro.compression.codec=deflate',
           '--conf', 'spark.hadoop.mapreduce.fileoutputcommitter.marksuccessfuljobs=false',
           jar]
    cmd.extend(get_param_list(params))
    return cmd


def run_cmd(cmd):
    """ run gdmix job locally.
    Params:
        cmd: shell command, e.g. ['spark-submit', '--class', ...]
    """
    process = Popen(cmd, stdout=PIPE, stderr=PIPE)
    # wait for the process to terminate
    out, err = process.communicate()
    returnCode = process.returncode
    print(out.decode("utf-8"))
    if returnCode != 0:
        raise RuntimeError("ERROR in executing commnd: {}\n\nError message:\n{}".format(
            str(' '.join(cmd)), err.decode("utf-8")))
    else:
        print(err.decode("utf-8"))



from gdmixworkflow.common.utils import json_config_file_to_obj, gen_random_string
from gdmixworkflow.common.constants import *
from gdmixworkflow.distributed.container_ops import no_op
from gdmixworkflow.fixed_effect_workflow_generator \
    import FixedEffectWorkflowGenerator
from gdmixworkflow.random_effect_workflow_generator \
    import RandomEffectWorkflowGenerator
import kfp.dsl as dsl


@dsl.pipeline()
def gdmix_distributed_workflow(gdmix_config_file, namespace, secret_name, image, service_account):
    """ Generate gdmix kubeflow pipeline using Kubeflow pipeline python DSL( kfp.dsl).
    """

    gdmix_config_obj = json_config_file_to_obj(gdmix_config_file)

    current_op = no_op("GDMix-training-start")
    suffix = gen_random_string()

    if hasattr(gdmix_config_obj, FIXED_EFFECT_CONFIG):
        fe_tip_op = no_op("fixed-effect-training-start")
        fe_tip_op.after(current_op)
        fe_workflow = FixedEffectWorkflowGenerator(gdmix_config_obj,
                                                   namespace=namespace,
                                                   secret_name=secret_name,
                                                   image=image,
                                                   service_account=service_account,
                                                   job_suffix=suffix)
        fe_start_op, current_op = fe_workflow.gen_workflow()
        fe_start_op.after(fe_tip_op)

    if hasattr(gdmix_config_obj, RANDOM_EFFECT_CONFIG):
        re_tip_op = no_op("random-effect-training-start")
        re_tip_op.after(current_op)
        re_workflow = RandomEffectWorkflowGenerator(gdmix_config_obj,
                                                   namespace=namespace,
                                                   secret_name=secret_name,
                                                   image=image,
                                                   service_account=service_account,
                                                   job_suffix=suffix)
        re_start_op, _ = re_workflow.gen_workflow()
        re_start_op.after(re_tip_op)


from gdmix.util.constants import *

DISTRIBUTED = "distributed"
FIXED_EFFECT_CONFIG = "fixed_effect_config"
GDMIX_TFJOB = "gdmix_tfjob"
GDMIX_SPARKJOB = "gdmix_sparkjob"
METRIC = "metric"
MODELS = "models"
RANDOM_EFFECT_CONFIG = "random_effect_config"
SINGLE_NODE = "single_node"
TRAINING_SCORES = "training_scores"
VALIDATION_SCORES = "validation_scores"
DETEXT_MODEL_OUTPUT_DIR = "out_dir"
DETEXT_DEV_FILE = "dev_file"
DETEXT_TRAIN_FILE = "train_file"



from gdmixworkflow.common.utils import *
from gdmixworkflow.common.constants import *
from gdmixworkflow.fixed_effect_workflow_generator \
    import FixedEffectWorkflowGenerator
from gdmixworkflow.random_effect_workflow_generator \
    import RandomEffectWorkflowGenerator
import json
import os
from os.path import join as path_join
import shutil


def create_output_dirs(gdmix_config_obj):
    """ Create output directories """

    def create_subdirs(root_dir):
        if os.path.isdir(root_dir):
            shutil.rmtree(root_dir)
        os.makedirs(root_dir)
        for sub_dir_name in [MODELS, METRIC, TRAINING_SCORES, VALIDATION_SCORES]:
            sub_dir = path_join(root_dir, sub_dir_name)
            os.makedirs(sub_dir)

    output_dir = gdmix_config_obj.output_dir
    if hasattr(gdmix_config_obj, FIXED_EFFECT_CONFIG):
        root_dir = path_join(output_dir, gdmix_config_obj.fixed_effect_config.name)
        create_subdirs(root_dir)

    if hasattr(gdmix_config_obj, RANDOM_EFFECT_CONFIG):
        for re_config in gdmix_config_obj.random_effect_config:
            root_dir = path_join(output_dir, re_config.name)
            create_subdirs(root_dir)

            num_partitions = re_config.num_partitions
            for score_output_name in [TRAINING_SCORES, VALIDATION_SCORES]:
                sub_dir = path_join(root_dir, score_output_name)
                for idx in range(num_partitions):
                    os.makedirs(path_join(sub_dir, "partitionId={}".format(idx)))


def run_gdmix_single_node(gdmix_config_file, jar_path):
    """ Run gdmix jobs locally including:
        - fixed-effect jobs
        - random-effect jobs
    """
    gdmix_config_obj = json_config_file_to_obj(gdmix_config_file)
    output_dir = gdmix_config_obj.output_dir

    create_output_dirs(gdmix_config_obj)

    if hasattr(gdmix_config_obj, FIXED_EFFECT_CONFIG):
        fe_workflow = FixedEffectWorkflowGenerator(gdmix_config_obj, jar_path=jar_path)
        fe_workflow.run()

    if hasattr(gdmix_config_obj, RANDOM_EFFECT_CONFIG):
        re_workflow = RandomEffectWorkflowGenerator(gdmix_config_obj, jar_path=jar_path)
        re_workflow.run()

    return output_dir


import argparse
from functools import partial, update_wrapper
from gdmixworkflow.common.constants import SINGLE_NODE, DISTRIBUTED
from gdmixworkflow.distributed_workflow import gdmix_distributed_workflow
from gdmixworkflow.single_node_workflow import run_gdmix_single_node
import os
import sys


def str2bool(v):
    """
    handle argparse can't parse boolean well.
    ref: https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse/36031646
    """
    if isinstance(v, bool):
        return v
    if isinstance(v, str):
        return v.lower() == 'true'
    else:
        raise argparse.ArgumentTypeError('Boolean or string value expected.')


def get_parser():
    """ Creates an argument parser.  """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config_path',
        type=str,
        required=True,
        help='path to gdmix config')
    parser.add_argument(
        '--mode',
        type=str,
        default=SINGLE_NODE,
        help='distributed or single_node')
    parser.add_argument(
        '--jar_path',
        type=str,
        default="gdmix-data-all_2.11.jar",
        help='local path to the gdmix-data jar for GDMix processing'
        'intermediate data, single_node only')
    parser.add_argument(
        '--workflow_name',
        type=str,
        default="gdmix-workflow",
        help='name for the generated zip file to upload to'
        'Kubeflow Pipeline, distributed mode only')
    parser.add_argument(
        '--namespace',
        type=str,
        default="default",
        help='Kubernetes namespace, distributed mode only')
    parser.add_argument(
        '--secret_name',
        type=str,
        default="default",
        help='secret name to access storage, distributed mode only')
    parser.add_argument(
        '--image',
        type=str,
        default="linkedin/gdmix",
        help='image used to launch gdmix jobs on Kubernetes, '
        'distributed mode only')
    parser.add_argument(
        '--service_account',
        type=str,
        default="default",
        help='service account to launch spark job, distributed mode only')
    return parser


def main(args=None):
    parser = get_parser()
    args = parser.parse_args(args)

    if args.mode == SINGLE_NODE:
        try:
            outputDir = run_gdmix_single_node(args.config_path, args.jar_path)
        except RuntimeError as err:
            print(str(err))
            sys.exit(1)

        print("""
------------------------
GDMix training is finished, results are saved to {}.
            """.format(outputDir))

    elif args.mode == DISTRIBUTED:
        if not args.namespace:
            print("ERROR: --namespace is required for distributed mode")
            sys.exit(1)

        def wrapped_partial(func, *args, **kwargs):
            partial_func = partial(func, *args, **kwargs)
            update_wrapper(partial_func, func)
            return partial_func

        func = wrapped_partial(
            gdmix_distributed_workflow,
            args.config_path,
            args.namespace,
            args.secret_name,
            args.image,
            args.service_account)

        outputFileName = args.workflow_name + ".zip"

        import kfp.compiler as compiler
        compiler.Compiler().compile(func, outputFileName)
        print("Workflow file is saved to {}".format(outputFileName))

    else:
        print("ERROR: --mode={} isn't supported.".format(args.mode))
        sys.exit(1)


if __name__ == "__main__":
    main()



from abc import abstractmethod
from gdmixworkflow.common.constants import *
from gdmixworkflow.common.utils import rm_backslash, join_params
from gdmixworkflow.distributed.container_ops import gdmix_tfjob_op, gdmix_sparkjob_op
from gdmixworkflow.single_node.local_ops import get_tfjob_cmd, get_sparkjob_cmd, run_cmd
from os.path import join as path_join


class WorkflowGenerator(object):
    def __init__(self, gdmix_config_obj, jar_path="", namespace="",
                 secret_name="", image="", service_account="", job_suffix=""):
        """ Init to generate full gdmix fix-effect or random-effect workflow.
            the config obj is a named tuple and its name/value pair can be passed
            as parameters to tensorflow jobs, need to handle itermediate and output paths
        """
        self.gdmix_config_obj = gdmix_config_obj
        self.namespace = namespace
        self.secret_name = secret_name
        self.image = image
        self.service_account = service_account
        self.suffix = job_suffix
        self.jar_path = jar_path

    def tip(self, job_name, cmd):
        """ single-node: show message for current job.
        """
        print("""
------------------------
    {}
------------------------
              """.format(self.get_name(job_name)))
        print("Executing cmd:\n  {}\n".format(' '.join(cmd)))

    def get_name(self, name):
        """ Append suffix to name and transform to lowercase. k8s requires
        job name must be lower case alphanumeric characters, '-' or '.', and
        full job name length should < 64. Here limit to 40 for potential
        suffix for subsequent jobs.
        """
        LENGTH = 40
        full_name = "{}-{}".format(name, self.suffix) if self.suffix else name
        return full_name.lower()[:LENGTH]

    def get_tfjob_config(self, extra_config):
        """ get tf job config for tf job container. """
        tfjob_resource_config = self.gdmix_config_obj.tfjob_config
        tfjob_config = {
            "namespace": self.namespace,
            "secretName": self.secret_name,
            "image": self.image,
            "workerType": tfjob_resource_config.worker_type,
            "memorySize": tfjob_resource_config.memory_size,
            "needChief": tfjob_resource_config.need_chief,
            "psNum": tfjob_resource_config.num_ps,
            "evaluatorNum": tfjob_resource_config.num_evaluator,
            "workerNum": tfjob_resource_config.num_worker
        }

        tfjob_config.update(extra_config)
        return tfjob_config

    def get_sparkjob_config(self, extra_config):
        """ get spark job config for spark job container. """
        spark_resource_config = self.gdmix_config_obj.spark_config
        spark_job_config = {
            "namespace": self.namespace,
            "secretName": self.secret_name,
            "image": self.image,
            "serviceAccount": self.service_account,
            "driverMemory": spark_resource_config.driver_memory_size,
            "executorCores": spark_resource_config.num_executor_core,
            "executorInstances": spark_resource_config.num_cxecutor,
            "executorMemory": spark_resource_config.executor_memory_size
        }
        spark_job_config.update(extra_config)
        return spark_job_config

    @abstractmethod
    def get_job_sequence(self):
        pass

    def gen_workflow(self):
        """ Generate gdmix fixed/random effect workflow from given job sequence.
            Returns:
                start_op: first job of the workflow
                end_op: last job of the workflow
                trainScorePath: training data score result path
                validationScorePath: validation data score result path
        """
        jobs = self.get_job_sequence()

        start_op = prev_op = current_op = None
        for (job_type, job_name, class_name, params) in jobs:
            if job_type == GDMIX_SPARKJOB:
                extra_config = {
                    "name": self.get_name(job_name),
                    "mainClass": class_name,
                    "arguments": join_params(params)
                }
                sparkjob_config = self.get_sparkjob_config(extra_config)
                current_op = gdmix_sparkjob_op(**sparkjob_config)
            elif job_type == GDMIX_TFJOB:
                cmd = "python -m gdmix.gdmix {}".format(join_params(params))
                extra_config = {
                    "name": self.get_name(job_name),
                    "cmd": cmd,
                }
                tfjob_config = self.get_tfjob_config(extra_config)
                current_op = gdmix_tfjob_op(**tfjob_config)

            if not prev_op:
                start_op = prev_op = current_op
            else:
                current_op.after(prev_op)
                prev_op = current_op

        return (start_op, current_op)

    def run(self):
        """ Run all gdmix fixed/random effect jobs at local from given job sequence.
        """
        jobs = self.get_job_sequence()

        for (job_type, job_name, class_name, params) in jobs:
            if job_type == GDMIX_SPARKJOB:
                cmd = get_sparkjob_cmd(class_name, rm_backslash(params), self.jar_path)
            elif job_type == GDMIX_TFJOB:
                cmd = get_tfjob_cmd(params)
            self.tip(job_name, cmd)
            run_cmd(cmd)


from copy import deepcopy
from gdmixworkflow.common.utils import json_config_file_to_obj
from gdmixworkflow.fixed_effect_workflow_generator \
    import FixedEffectWorkflowGenerator
from gdmixworkflow.random_effect_workflow_generator \
    import RandomEffectWorkflowGenerator
import json
import os
from os.path import join as path_join
import unittest

class TestGDMixWorkflowGenerator(unittest.TestCase):
    """
    Test gdmix workflow generator
    """

    def setUp(self):
        lr_config_file = path_join(
            os.getcwd(),
            "test/resources/lr-single-node-movieLens.config")
        self.lr_config_obj = json_config_file_to_obj(lr_config_file)

        detext_config_file = path_join(
            os.getcwd(),
            "test/resources/detext-single-node-movieLens.config")
        self.detext_config_obj = json_config_file_to_obj(detext_config_file)

        # Set self.maxDiff to None to see diff for long text
        self.maxDiff = None

    def test_lr_model_fixed_effect_workflow_generator(self):
        fe_workflow = FixedEffectWorkflowGenerator(self.lr_config_obj)
        # check sequence
        seq = fe_workflow.get_job_sequence()
        self.assertEqual(len(seq), 2)
        # check job properties
        actual_train_job = seq[0]
        actual_compute_metric_job = seq[1]
        expected_train_job = (
            'gdmix_tfjob',
            'global-tf-train',
            '',
            {'--stage': 'fixed_effect',
             '--action': 'train',
             '--model_type': 'logistic_regression',
             '--training_data_dir': 'movieLens/global/trainingData',
             '--validation_data_dir': 'movieLens/global/validationData',
             '--copy_to_local': False,
             '--feature_file': 'movieLens/global/featureList/global',
             '--label_column_name': 'response',
             '--uid_column_name': 'uid',
             '--weight_column_name': 'weight',
             '--feature_bag': 'global',
             '--metadata_file': 'movieLens/global/metadata/tensor_metadata.json',
             '--l2_reg_weight': 1.0,
             '--regularize_bias': False,
             '--optimizer_name': 'LBFGS',
             '--lbfgs_tolerance': 1e-12,
             '--num_of_lbfgs_iterations': 100,
             '--num_of_lbfgs_curvature_pairs': 10,
             '--prediction_score_column_name': 'predictionScore',
             '--output_model_dir': 'lr-training/global/models',
             '--training_score_dir': 'lr-training/global/training_scores',
             '--validation_score_dir': 'lr-training/global/validation_scores'})

        expected_compute_metric_job = (
            'gdmix_sparkjob',
            'global-compute-metric',
            'com.linkedin.gdmix.evaluation.AreaUnderROCCurveEvaluator',
            {'\\--metricsInputDir': 'lr-training/global/validation_scores',
             '--outputMetricFile': 'lr-training/global/metric',
             '--labelColumnName': 'response',
             '--predictionColumnName': 'predictionScore'})
        self.assertEqual(actual_train_job, expected_train_job)
        self.assertEqual(actual_compute_metric_job, expected_compute_metric_job)

    def test_detext_model_fixed_effect_workflow_generator(self):
        fe_workflow = FixedEffectWorkflowGenerator(self.detext_config_obj)
        # check sequence
        seq = fe_workflow.get_job_sequence()
        self.assertEqual(len(seq), 3)
        # check job properties
        actual_train_job = seq[0]
        actual_inference_job = seq[1]
        actual_compute_metric_job = seq[2]

        expected_train_job_param = {
            '--stage': 'fixed_effect',
            '--model_type': 'detext',
            '--ftr_ext': 'cnn',
            '--elem_rescale': True,
            '--ltr_loss_fn': 'pointwise',
            '--init_weight': '0.1',
            '--learning_rate': '0.002',
            '--num_classes': '1',
            '--max_len': '16',
            '--min_len': '3',
            '--num_filters': '50',
            '--num_train_steps': '1000',
            '--num_units': '64',
            '--optimizer': 'bert_adam',
            '--pmetric': 'auc',
            '--all_metric': 'auc',
            '--steps_per_stats': '10',
            '--steps_per_eval': '100',
            '--train_batch_size': '64',
            '--test_batch_size': '64',
            '--use_deep': True,
            '--vocab_file': 'movieLens/detext/vocab.txt',
            '--resume_training': False,
            '--feature_names': 'label,doc_query,uid,wide_ftrs_sp_idx,wide_ftrs_sp_val',
            '--num_wide_sp': '45',
            '--train_file': 'movieLens/detext/trainingData/train_data.tfrecord',
            '--dev_file': 'movieLens/detext/validationData/test_data.tfrecord',
            '--test_file': 'movieLens/detext/validationData/test_data.tfrecord',
            '--metadata_file': 'movieLens/per-user/metadata/tensor_metadata.json',
            '--label_column_name': 'response',
            '--uid_column_name': 'uid',
            '--weight_column_name': 'weight',
            '--feature_bag': 'global',
            '--keep_checkpoint_max': '1',
            '--prediction_score_column_name': 'predictionScore',
            '--out_dir': 'detext-training/global/models'}
        expected_train_job = ('gdmix_tfjob', 'global-tf-train', '', expected_train_job_param)
        self.assertEqual(expected_train_job, actual_train_job)

        expected_inference_job_param = deepcopy(expected_train_job_param)
        expected_inference_job_param["--training_data_dir"] = 'movieLens/detext/trainingData/train_data.tfrecord'
        expected_inference_job_param["--training_score_dir"] = "detext-training/global/training_scores"
        expected_inference_job_param["--validation_data_dir"] = 'movieLens/detext/validationData/test_data.tfrecord'
        expected_inference_job_param["--validation_score_dir"] = "detext-training/global/validation_scores"
        expected_inference_job_param["--action"] = "inference"
        expected_inference_job = (
            'gdmix_tfjob',
            'global-tf-inference',
            '',
            expected_inference_job_param)
        self.assertEqual(expected_inference_job, actual_inference_job)

        expected_compute_metric_job = (
            'gdmix_sparkjob',
            'global-compute-metric',
            'com.linkedin.gdmix.evaluation.AreaUnderROCCurveEvaluator',
            {'\\--metricsInputDir': 'detext-training/global/validation_scores',
             '--outputMetricFile': 'detext-training/global/metric',
             '--labelColumnName': 'response',
             '--predictionColumnName': 'predictionScore'})
        self.assertEqual(actual_compute_metric_job, expected_compute_metric_job)

    def test_lr_model_random_effect_workflow_generator(self):
        re_workflow = RandomEffectWorkflowGenerator(self.lr_config_obj)
        # check sequence
        seq = re_workflow.get_job_sequence()
        self.assertEqual(len(seq), 6)
        # check job properties
        actual_partition_job = seq[0]
        actual_train_job = seq[1]
        actual_compute_metric_job = seq[2]

        expected_partition_job = (
            'gdmix_sparkjob',
            'per-user-partition',
            'com.linkedin.gdmix.data.DataPartitioner',
            {'\\--trainingDataDir': 'movieLens/per_user/trainingData',
             '--validationDataDir': 'movieLens/per_user/validationData',
             '--metadataFile': 'movieLens/per_user/metadata/tensor_metadata.json',
             '--partitionId': 'user_id',
             '--numPartitions': 1,
             '--dataFormat': 'tfrecord',
             '--partitionedTrainingDataDir': 'lr-training/per-user/partition/trainingData',
             '--partitionedValidationDataDir': 'lr-training/per-user/partition/validationData',
             '--outputMetadataFile': 'lr-training/per-user/partition/metadata/tensor_metadata.json',
             '--outputPartitionListFile': 'lr-training/per-user/partition/partitionList.txt',
             '--predictionScoreColumnName': 'predictionScore',
             '--trainingScoreDir': 'lr-training/global/training_scores',
             '--validationScoreDir': 'lr-training/global/validation_scores'})

        expected_train_job = (
            'gdmix_tfjob',
            'per-user-tf-train',
            '',
            {'--stage': 'random_effect',
             '--action': 'train',
             '--model_type': 'logistic_regression',
             '--partition_entity': 'user_id',
             '--training_data_dir': 'lr-training/per-user/partition/trainingData',
             '--validation_data_dir': 'lr-training/per-user/partition/validationData',
             '--feature_file': 'movieLens/per_user/featureList/per_user',
             '--label_column_name': 'response',
             '--uid_column_name': 'uid',
             '--weight_column_name': 'weight',
             '--feature_bag': 'per_user',
             '--metadata_file': 'lr-training/per-user/partition/metadata/tensor_metadata.json',
             '--l2_reg_weight': 1.0,
             '--regularize_bias': False,
             '--num_partitions': 1,
             '--optimizer_name': 'LBFGS',
             '--lbfgs_tolerance': 1e-12,
             '--num_of_lbfgs_iterations': 100,
             '--num_of_lbfgs_curvature_pairs': 10,
             '--max_training_queue_size': 10,
             '--num_of_consumers': 1,
             '--enable_local_indexing': False,
             '--prediction_score_column_name': 'predictionScore',
             '--partition_list_file': 'lr-training/per-user/partition/partitionList.txt',
             '--output_model_dir': 'lr-training/per-user/models',
             '--training_score_dir': 'lr-training/per-user/training_scores',
             '--validation_score_dir': 'lr-training/per-user/validation_scores'})

        expected_compute_metric_job = (
            'gdmix_sparkjob',
            'per-user-compute-metric',
            'com.linkedin.gdmix.evaluation.AreaUnderROCCurveEvaluator',
            {'\\--metricsInputDir': 'lr-training/per-user/validation_scores',
             '--outputMetricFile': 'lr-training/per-user/metric',
             '--labelColumnName': 'response',
             '--predictionColumnName': 'predictionScore'})

        self.assertEqual(expected_partition_job, actual_partition_job)
        self.assertEqual(expected_train_job, actual_train_job)
        self.assertEqual(expected_compute_metric_job, actual_compute_metric_job)


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple
from gdmixworkflow.common.constants import *
from itertools import product
import json
from subprocess import Popen,PIPE,STDOUT


def gen_random_string(length=6):
    """ Generate fixed-length random string. """
    import random
    import string
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


def abbr(name):
    """ Return abbreviation of a given name.
    Example:
        fixed-effect -> f10t
        per-member -> p8r
    """
    return "{}{}{}".format(name[0], len(name) - 2,
                           name[-1]) if len(name) > 2 else name


def prefix_dash_dash(params):
    """ Add -- for keys in gdmix tfjob params. """
    if isinstance(params, dict):
        newParams = {}
        for k, v in params.items():
            newParams["--{}".format(k)] = v
        return newParams
    else:
        raise ValueError("job params can only be dict")


def join_params(params):
    """ Join param to string as key value pair. If the key begins
    with '#', the key is ignored.
    """
    if isinstance(params, dict):
        return (' ').join(["{} {}".format(k, v) if not k.startswith(
            '#') else str(v) for (k, v) in params.items()])
    else:
        raise ValueError("job params can only be dict")


def rm_backslash(params):
    """ A '-' at the begining of a line is a special charter in YAML,
    used backslash to escape, need to remove the added backslach for local run.
    """
    newParams = {}
    for k, v in params.items():
        newParams[k.strip('\\')] = v
    return newParams


def json_config_file_to_obj(config_file):
    """ load gdmix config from json file to object. """
    def _json_object_hook(d):
        return namedtuple('GDMIX_CONFIG', d.keys())(*d.values())

    with open(config_file) as f:
        config_obj = json.load(f, object_hook=_json_object_hook)
    return config_obj


def flatten_config_obj(d, config_obj):
        """ flatten a config obj to a dict without nested dict.
        For example: {a: {b: c}} --> {b: c}
        """
        for k, v in config_obj._asdict().items():
            if type(v) not in [str, bool, int, float, list]:
                flatten_config_obj(d, v)
            else:
                d[k] = v
        return d


from gdmixworkflow.common.utils import *
from gdmixworkflow.common.constants import *
from gdmixworkflow.workflow_generator import WorkflowGenerator
import os
from os.path import join as path_join


class RandomEffectWorkflowGenerator(WorkflowGenerator):
    """ Generate gdmix random effect workflow consisting of
    - sparkjob: partition
    - tfjob: train
    - sparkjob: compute-metric
    """

    def __init__(self, gdmix_config_obj, jar_path="", namespace="",
                 secret_name="", image="", service_account="", job_suffix=""):
        """ Init to generate gdmix random effect workflow. """
        super().__init__(gdmix_config_obj, jar_path, namespace, secret_name, image, service_account, job_suffix)
        self.root_output_dir = gdmix_config_obj.output_dir

    def get_prev_model_score_paths(self, prev_model_name):
        """ Get paths of the scoring results from fixed-effect model or random-effect
        model before it. """
        output_dir = path_join(self.root_output_dir, prev_model_name)
        prev_train_score_dir = path_join(output_dir, TRAINING_SCORES)
        prev_validation_score_dir = path_join(output_dir, VALIDATION_SCORES)
        return (prev_train_score_dir, prev_validation_score_dir)

    def get_train_output_paths(self, random_effect_name):
        """ Get output paths for the training job. """
        output_dir = path_join(self.root_output_dir, random_effect_name)
        output_model_dir = path_join(output_dir, MODELS)
        training_score_dir = path_join(output_dir, TRAINING_SCORES)
        validation_score_dir = path_join(output_dir, VALIDATION_SCORES)

        return (output_model_dir, training_score_dir, validation_score_dir)

    def get_train_input_paths(self, random_effect_name):
        """ Get input paths for the training job that are from the partition job. """
        output_dir = path_join(self.root_output_dir,
                              path_join(random_effect_name, "partition"))
        training_data_dir = path_join(output_dir, "trainingData")
        validation_data_dir = path_join(output_dir, "validationData")
        metadata_file = path_join(output_dir,
                                 path_join("metadata","tensor_metadata.json"))
        partition_list_file = path_join(output_dir, "partitionList.txt")
        return (training_data_dir, validation_data_dir, metadata_file, partition_list_file)

    def get_metric_output_path(self, random_effect_name):
        """ Get metric output path. """
        output_dir = path_join(self.root_output_dir, random_effect_name)
        metric_output_path = path_join(output_dir, METRIC)
        return metric_output_path

    def get_partition_job(self, random_effect_config_obj, prev_model_name):
        """ Get sparkjob partition command.
        Return: [job_type, job_name, "", job_params]
        """
        prev_train_score_dir, prev_validation_score_dir = \
        self.get_prev_model_score_paths(prev_model_name)

        random_effect_name = random_effect_config_obj.name
        training_data_dir, validation_data_dir, metadata_file, partition_list_file = \
        self.get_train_input_paths(random_effect_name)

        params = {
            r"\--trainingDataDir": random_effect_config_obj.training_data_dir,
            "--validationDataDir": random_effect_config_obj.validation_data_dir,
            "--metadataFile": random_effect_config_obj.metadata_file,
            "--partitionId": random_effect_config_obj.partition_entity,
            "--numPartitions": random_effect_config_obj.num_partitions,
            "--dataFormat": TFRECORD,
            "--partitionedTrainingDataDir": training_data_dir,
            "--partitionedValidationDataDir": validation_data_dir,
            "--outputMetadataFile": metadata_file,
            "--outputPartitionListFile": partition_list_file,
            "--predictionScoreColumnName": random_effect_config_obj.prediction_score_column_name,
            "--trainingScoreDir": prev_train_score_dir,
            "--validationScoreDir": prev_validation_score_dir
        }
        return (GDMIX_SPARKJOB, "{}-partition".format(random_effect_name),
                "com.linkedin.gdmix.data.DataPartitioner", params)

    def get_train_job(self, random_effect_config_obj):
        """ Get tfjob training job.
        Return: [job_type, job_name, "", job_params]
        """
        random_effect_name = random_effect_config_obj.name
        params = {STAGE: RANDOM_EFFECT}
        # get params from config
        flatten_config_obj(params, random_effect_config_obj)
        # adjust param keys
        params.pop("name")
        params[PREDICTION_SCORE_COLUMN_NAME] = params.pop(PREDICTION_SCORE_COLUMN_NAME)
        # adjust training/validation data and metadata path as the output of partition job
        training_data_dir, validation_data_dir, metadata_file, partition_list_file = \
        self.get_train_input_paths(random_effect_name)
        params[TRAINING_DATA_DIR] = training_data_dir
        params[VALIDATION_DATA_DIR] = validation_data_dir
        params[METADATA_FILE] = metadata_file
        params[PARTITION_LIST_FILE] = partition_list_file
        # add output params
        output_model_dir, training_score_dir, validation_score_dir = \
        self.get_train_output_paths(random_effect_name)

        params[OUTPUT_MODEL_DIR] = output_model_dir
        params[TRAINING_SCORE_DIR] = training_score_dir
        params[VALIDATION_SCORE_DIR] = validation_score_dir

        params = prefix_dash_dash(params)
        return (GDMIX_TFJOB, "{}-tf-train".format(random_effect_name), "", params)

    def get_compute_metric_job(self, random_effect_config_obj):
        """ Get sparkjob compute metrics command.
        Return: [job_type, job_name, class_name, job_params]
        """
        random_effect_name = random_effect_config_obj.name
        _, _, validation_score_dir = self.get_train_output_paths(random_effect_name)
        params = {
            r"\--metricsInputDir": validation_score_dir,
            "--outputMetricFile": self.get_metric_output_path(random_effect_name),
            "--labelColumnName": random_effect_config_obj.input_column_names.label_column_name,
            "--predictionColumnName": random_effect_config_obj.prediction_score_column_name
        }
        return (GDMIX_SPARKJOB, "{}-compute-metric".format(random_effect_name),
                "com.linkedin.gdmix.evaluation.AreaUnderROCCurveEvaluator",
                params)

    def get_job_sequence(self):
        """ Get gdmix job sequence.
        """
        random_effect_config_obj_list = self.gdmix_config_obj.random_effect_config
        jobs = []
        prev_model_name = self.gdmix_config_obj.fixed_effect_config.name

        for re_config_obj in random_effect_config_obj_list:
            jobs.append(self.get_partition_job(re_config_obj, prev_model_name))
            jobs.append(self.get_train_job(re_config_obj))
            jobs.append(self.get_compute_metric_job(re_config_obj))
            prev_model_name = re_config_obj.name
        return jobs



from copy import deepcopy
from os.path import join as path_join
import os
from gdmixworkflow.workflow_generator import WorkflowGenerator
from gdmixworkflow.common.utils import *
from gdmixworkflow.common.constants import *


class FixedEffectWorkflowGenerator(WorkflowGenerator):
    """ Generate gdmix fixed effect workflow consisting of
    - tfjob: train and inference training and validation data
    - sparkjob: compute-metric
    """

    def __init__(self, gdmix_config_obj, jar_path="", namespace="",
                 secret_name="", image="", service_account="", job_suffix=""):
        """ Init to generate gdmix fixed effect workflow. """
        super().__init__(gdmix_config_obj, jar_path, namespace, secret_name, image, service_account, job_suffix)
        self.fixed_effect_config_obj = self.gdmix_config_obj.fixed_effect_config
        self.fixed_effect_name = self.fixed_effect_config_obj.name
        self.output_dir = path_join(gdmix_config_obj.output_dir,
                                   self.fixed_effect_config_obj.name)
        self.output_model_dir = path_join(self.output_dir, MODELS)
        self.training_score_dir = path_join(self.output_dir, TRAINING_SCORES)
        self.validation_score_dir = path_join(self.output_dir, VALIDATION_SCORES)
        self.metric_file = path_join(self.output_dir, METRIC)
        self.model_type = self.fixed_effect_config_obj.model_type

    def get_train_job(self):
        """ Get tfjob training job.
        Return: (job_type, job_name, "", job_params)
        """
        params = {STAGE: FIXED_EFFECT}
        # get params from config
        flatten_config_obj(params, self.fixed_effect_config_obj)
        # adjust param keys
        params.pop("name")
        params[PREDICTION_SCORE_COLUMN_NAME] = params.pop(PREDICTION_SCORE_COLUMN_NAME)
        # add output params
        if self.model_type == LOGISTIC_REGRESSION:
            params[OUTPUT_MODEL_DIR] = self.output_model_dir
            params[TRAINING_SCORE_DIR] = self.training_score_dir
            params[VALIDATION_SCORE_DIR] = self.validation_score_dir
        elif self.model_type == DETEXT:
            params[DETEXT_MODEL_OUTPUT_DIR] = self.output_model_dir
        else:
            raise ValueError('unsupported model_type: {}'.format(model_type))

        params = prefix_dash_dash(params)
        return (GDMIX_TFJOB, "{}-tf-train".format(self.fixed_effect_name), "", params)

    def get_detext_inference_job(self):
        """ Get detext inference job. For LR model the inference job is included in train
        job, this job is for DeText model inference.
        Return: an inferece job inferencing training and validation data
        (job_type, job_name, "", job_params)
        """
        params = {STAGE: FIXED_EFFECT, ACTION: ACTION_INFERENCE}
        # get params from config
        flatten_config_obj(params, self.fixed_effect_config_obj)
        # adjust param keys
        params.pop("name")
        params[PREDICTION_SCORE_COLUMN_NAME] = params.pop(PREDICTION_SCORE_COLUMN_NAME)
        params[DETEXT_MODEL_OUTPUT_DIR] = self.output_model_dir
        # "--dev_file" and "--validation_output_dir" are used as input and output for the detext inference job
        inference_params = deepcopy(params)
        inference_params[TRAINING_DATA_DIR] = self.fixed_effect_config_obj.train_file
        inference_params[TRAINING_SCORE_DIR] = self.training_score_dir
        inference_params[VALIDATION_DATA_DIR] = self.fixed_effect_config_obj.dev_file
        inference_params[VALIDATION_SCORE_DIR] = self.validation_score_dir
        inference_job = (GDMIX_TFJOB, "{}-tf-inference".format(
            self.fixed_effect_name), "", prefix_dash_dash(inference_params))

        return inference_job

    def get_compute_metric_job(self):
        """ Get sparkjob compute metric job.
        Return: (job_type, job_name, class_name, job_params)
        """
        params = {
            r"\--metricsInputDir": self.validation_score_dir,
            "--outputMetricFile": self.metric_file,
            "--labelColumnName": self.fixed_effect_config_obj.input_column_names.label_column_name,
            "--predictionColumnName": self.fixed_effect_config_obj.prediction_score_column_name
        }
        return (GDMIX_SPARKJOB,
                "{}-compute-metric".format(self.fixed_effect_name),
                "com.linkedin.gdmix.evaluation.AreaUnderROCCurveEvaluator",
                params)

    def get_job_sequence(self):
        """ Get job sequence of fixed effect workflow. """
        if self.model_type == LOGISTIC_REGRESSION:
            jobs = [self.get_train_job(), self.get_compute_metric_job()]
        elif self.model_type == DETEXT:
            jobs = [self.get_train_job(),
                    self.get_detext_inference_job(),
                    self.get_compute_metric_job()]
        else:
            raise ValueError('unsupported model_type: {}'.format(model_type))
        return jobs



from pathlib import Path
from setuptools import find_namespace_packages, setup
from sys import platform as _platform

import sys


VERSION="0.3.0"
current_dir = Path(__file__).resolve().parent
with open(current_dir.joinpath('README.md'), encoding='utf-8') as f:
    long_description = f.read()

if _platform not in ["linux", "linux2", "darwin"]:
    print("ERROR: platform {} isn't supported".format(_platform))
    sys.exit(1)

setup(
    name="gdmix-workflow",
    python_requires='>=3.7',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=["Programming Language :: Python :: 3.7",
                 "Intended Audience :: Science/Research",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved"],
    license='BSD-2-CLAUSE',
    version=f'{VERSION}',
    package_dir={'': 'src'},
    packages=find_namespace_packages(where='src'),
    package_data={'': ['*.yaml']},
    include_package_data=True,
    install_requires=[
        "setuptools>=41.0.0",
        f"gdmix-trainer=={VERSION}",
        "google-auth==1.21.1",
        "kfp==0.2.5"
    ],
    tests_require=['pytest']
)



# DO NOT COPY & PASTE THIS CODE!!!!!
#
# This is a special file only needed for "src/gdmixworkflow/__init__.py"
# to declare the "gdmixworkflow" package as a "namespace"
#
# All other "__init__.py" files can just be blank, or contain normal Python
# module code.
try:
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)



import fileinput
from gdmixworkflow.common.utils import gen_random_string
from gdmixworkflow.distributed import resource
import os
from os.path import join as path_join
import shutil


def load_launcher_from_file(src_file, name_placeholder, name):
    """ Load launcher's commponent yaml file and update launcher name """
    dest_file = shutil.copyfile(src_file, gen_random_string(20) + ".tmp.yaml")
    with fileinput.FileInput(dest_file, inplace=True) as f:
        for line in f:
            line = line.replace(name_placeholder, name)
            print(line, end='')
    from kfp import components
    component = components.load_component_from_file(dest_file)
    os.remove(dest_file)
    return component


def gdmix_tfjob_op(
        name,
        cmd,
        namespace,
        secretName,
        image="linkedin/gdmix",
        workerType="cpu",
        needChief=False,
        psNum=0,
        evaluatorNum=0,
        workerNum=2,
        memorySize='1G',
        ttlSecondsAfterFinished=-1,
        tfjobTimeoutMinutes=1440,
        deleteAfterDone=False):
    """
    This function prepares params for launch_tfjob.py, as defined in the
    templates of ../launcher/tfjob/tfjob_component.yaml, which specifies to
    execute the launch_tfjob.py once the container is ready. In the container,
    The launch_tfjob.py assemble all params to form a deployable YAML file to
    launch the actual TFJob.
    """
    componentPath = path_join(resource.__path__[0],
                             "tfjob_component.yaml")
    name_placeholder = "TFJob-launcher-name"
    tfjob_launcher_op = load_launcher_from_file(
        componentPath, name_placeholder, name)

    # Container spec of a TFJob worker
    containerSpec = {
        "command": ["sh", "-c", "{}".format(cmd)],
        "image": "{}".format(image),
        "resources": {"limits": {"memory": "{}i".format(memorySize.upper())},
                      "requests": {"memory": "{}i".format(memorySize.upper())}},
        "name": "tensorflow",
        "volumeMounts": [{"name": "shared-data",
                          "mountPath": "/var/tmp"},
                         {"name": "dt",
                          "mountPath": "/var/dt"}]
    }

    # Volume spec for model code and secrets
    volumeSpec = [{"name": "shared-data",
                   "emptyDir": {"sizeLimit": "10Mi"}
                  },
                  {"name": "dt",
                  "secret": {"secretName": secretName,
                             "defaultMode": 256}
                  }]

    if (workerType == "gpu"):
        containerSpec["resources"]["limits"]["nvidia.com/gpu"] = 1

    workerSpecTemplate = {
        "replicas": 1,
        "restartPolicy": "OnFailure",
        "tfReplicaType": "WORKER",
        "template": {
            "spec": {
                "containers": [containerSpec],
                "volumes": volumeSpec
            }
        }
    }

    chief = ps = evaluator = worker = {}

    if needChief:
        chief = workerSpecTemplate.copy()
        chief["tfReplicaType"] = "MASTER"
    if psNum > 0:
        ps = workerSpecTemplate.copy()
        ps["tfReplicaType"] = "PS"
        ps["replicas"] = psNum
    if evaluatorNum > 0:
        evaluator = workerSpecTemplate.copy()
        evaluator["tfReplicaType"] = "EVALUATOR"
        evaluator["replicas"] = evaluatorNum
    if workerNum > 0:
        worker = workerSpecTemplate.copy()
        worker["replicas"] = workerNum

    op = tfjob_launcher_op(
        name=name,
        namespace=namespace,
        ttl_seconds_after_finished=ttlSecondsAfterFinished,
        ps_spec=ps,
        worker_spec=worker,
        chief_spec=chief,
        evaluator_spec=evaluator,
        tfjob_timeout_minutes=tfjobTimeoutMinutes,
        delete_finished_tfjob=deleteAfterDone
    )
    return op


def gdmix_sparkjob_op(
        name,
        mainClass,
        arguments,
        secretName,
        mainApplicationFile="local:///opt/spark/jars/gdmix-data-all_2.11.jar",
        image="linkedin/gdmix",
        namespace="default",
        serviceAccount="default",
        driverCores=1,
        driverMemory='2g',
        executorCores=2,
        executorInstances=2,
        executorMemory='1g',
        sparkApplicationTimeoutMinutes=1440,
        deleteAfterDone=False):
    """
    This function prepares params for launch_sparkapplication.py as defined in
    the templates of ../launcher/sparkapplication/sparkapplication_component.yaml,
    which specifies to execute the launch_tfjob.py once the container is ready.
    In the container, the launch_tfjob.py assemble all params to form a deployable
    YAML file to launch the actual spark application.
    """
    componentPath = path_join(resource.__path__[0],
                             "sparkapplication_component.yaml")
    name_placeholder = "SparkApplication-launcher-name"
    spark_application_launcher_op = load_launcher_from_file(
        componentPath, name_placeholder, name)

    # Driver spec for the spark application
    # Note: secret for now is pre-craeted and will be valid for 7 days
    driverSpec = {
        "cores": driverCores,
        "memory": driverMemory,
        "secrets": [
            {
                "name": secretName,
                "path": "/var/tmp/{}".format(secretName),
                "secretType": "HadoopDelegationToken"
            }
        ],
        "serviceAccount": serviceAccount
    }

    # Executor spec for the spark application
    executorSpec = {
        "cores": executorCores,
        "instances": executorInstances,
        "memory": executorMemory
    }

    return spark_application_launcher_op(
        name=name,
        namespace=namespace,
        image=image,
        main_class=mainClass,
        arguments=arguments,
        main_application_file=mainApplicationFile,
        driver_spec=driverSpec,
        executor_spec=executorSpec,
        sparkapplication_timeout_minutes=sparkApplicationTimeoutMinutes,
        delete_finished_sparkapplication=deleteAfterDone
    )


def no_op(msg):
    """ Empty op as a placeholder on workflow to handle dependency.
    """
    from kfp import dsl
    return dsl.ContainerOp(
        name="{}-{}".format(msg, gen_random_string()),
        image="alpine",
        command=['sh', '-c'],
        arguments=["echo {}".format(msg)]
    )
